package com.sillycycle.bagleyd.life2d;

/*
 * @(#)Life2D.java
 *
 * Copyright 1993 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Cursor;
import java.awt.MenuItem;
import java.awt.Point;
import java.awt.PopupMenu;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
//import java.awt.event.MouseWheelEvent;
//import java.awt.event.MouseWheelListener;
import javax.swing.JFrame;

import com.sillycycle.bagleyd.util.ArgumentParser;

/**
 * The <code>Life2D</code> class extends the
 * <code>Life2DCanvas</code> for handling the responses to the user
 * for the <code>Life2DFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

class Life2D extends Life2DCanvas implements Runnable,
		MouseListener, MouseMotionListener, KeyListener {
	private static final long serialVersionUID = 42L;

	int now = 0;
	byte cellState = 0;
	boolean quit = false;
	int numberedSearch = -1;
	boolean fileSearch = false; // used?
	int fileSearchShake = -1;
	boolean patternSearch = false;
	boolean gliderSearch = false;
	String file0String = "";
	String file1String = "";
	String[] toggleStamp = new String[4]; // initialize "0", "1", "2", "3"
	private transient Thread gameRunner;
	PopupMenu popupMenu = null;
	transient Life2DPopup popupListener = null;

	static int X = 0;
	static int Y = 0;

	public Life2D(JFrame frame, int speed, int columns,
			int states, int neighbors,
			boolean randomOrient, boolean grid,
			boolean wrap, boolean visible, boolean mono,
			boolean round, boolean scroll, Color bg) {
		this.frame = frame;
		this.speed = speed;
		this.states = states; // to simulate B5/S678 with barriers
		this.neighbors = neighbors;
		this.ncols = getRealColumns(columns);
		this.nrows = getRowsFromColumns(columns);
		this.randomOrient = randomOrient;
		this.grid = grid;
		this.wrap = wrap;
		this.visible = visible;
		this.mono = mono;
		this.round = round;
		this.scroll = scroll;
		this.background = bg;
		setBackground(bg);
		if (frame != null) {
			// headless
			this.addKeyListener(this);
			this.addMouseListener(this);
			this.addMouseMotionListener(this);
			setCursor(new Cursor(Cursor.HAND_CURSOR));
			initPopup();
		}
		param = new Life2DRules(neighbors,
			Life2DRules.defaultRuleStrings[getNeighborIndex()]);
		inits();
		start();
	}

	public void setPopupMenu(PopupMenu popupMenu) {
		this.popupMenu = popupMenu;
		add(popupMenu);
	}

	@Override
	protected void processMouseEvent(MouseEvent event) {
		if ((popupMenu != null) && event.isPopupTrigger())
			popupMenu.show(this, event.getX(), event.getY());
		else
			super.processMouseEvent(event);
	}

	protected final void initPopup() {
		popupListener = new Life2DPopup(this);
		popupMenu = new PopupMenu("Select a template");

		MenuItem menuItem = new MenuItem("Cut");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Copy");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Paste");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Clear");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		popupMenu.addSeparator();

		menuItem = new MenuItem("Cw");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Ccw");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Refl Wall");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		menuItem = new MenuItem("Refl Grnd");
		menuItem.addActionListener(popupListener);
		popupMenu.add(menuItem);

		setPopupMenu(popupMenu);
	}

	void quitGame() {
		banner(Life2DInterface.MSG_END);
		quit = true;
	}

	void search() {
		if (patternSearch) {
			findLife(false);
		} else if (gliderSearch) {
			findLife(true);
		} else if (fileSearchShake != -1) {
			findFileLife(fileSearchShake);
		}
		toggleStamp[0] = "0";
		toggleStamp[1] = "1";
		toggleStamp[2] = "2";
		toggleStamp[3] = "3";
	}

	synchronized boolean processAdvance() {
		boolean gotNew = false;
		final int BIG = 5120; // give up after this

		if (!paused) {
			lifePulse(visible, speed == Life2DInterface.MAX_SPEED, true);
			if ((excludePeriod % 2 == 0 || excludePeriod % 3 == 0) && !findStable) {
				int modulus = gen % 4;
				int tripleOffset = (gen + 1) % 4;
				int toggleOffset = (gen + 2) % 4;

				toggleStamp[modulus] = getGridStamp();
				if ((toggleStamp[modulus].equals(toggleStamp[toggleOffset])) ||
						(toggleStamp[modulus].equals(toggleStamp[tripleOffset]))) {
					search();
					return gotNew;
				}
			}
			if (nCells <= 2) {
				search();
			} else if (gen + genOffset > searchLength) {
				// force true to print pattern
				if (patternSearch || gliderSearch) {
					printSearch(1);
					search();
				}
			} else {
				if (patternSearch || gliderSearch) {
					/* probably not useful as most oscillators
					   are too simple and not in junk list */
					/*if (gen % 16 == 0) {
						if (findPattern()) {
							search();
							return gotNew;
						}
					}*/
					if (!findStable && cCells == 0) {
						search();
					} else if (nCells > BIG) {
						if (unboundDetection) {
							System.out.println("Possibly unbounded");
							retrieveInitLife();
						}
						search();
					} else if (edge && edgeDetection) {
						printLife();
						search();
						/*if (genOffset == 0 && wrap) {
							if (!printSearch(0)) {
								search();
							} else {
								genOffset = searchLength - gen - excludePeriod + 1;
								if (genOffset < 0)
									genOffset = 0;
							}
						}*/
					} else if (checkFind == -1 &&
						gen + genOffset >= searchLength - excludePeriod + 1) {
						if (findPattern()) {
							search();
							return gotNew;
						}
						if (gen + genOffset == searchLength - excludePeriod + 1 &&
								!printSearch(0)) {
							search();
						}
					} else if (checkFind != -1 &&
							gen + genOffset >= searchLength - findExcludePeriod[checkFind] + 1) {
						if (findPattern()) {
							search();
							return gotNew;
						}
						if (gen + genOffset == searchLength - findExcludePeriod[checkFind] + 1 &&
								!printSearch(0)) {
							search();
						}
					}
				}
			}
		}
		return gotNew;
	}

	boolean processKey(char key) {
		char charPressed = key;

		switch (charPressed) {
		case '>':
		case '.':
			if (speed < Life2DInterface.MAX_SPEED) {
				speedUp();
			}
			return true;
		case '<':
		case ',':
			if (speed > Life2DInterface.MIN_SPEED) {
				slowDown();
			}
			return true;
		case '?':
			((Life2DFrame) frame).descriptionHelp();
			return true;
		case '!':
			((Life2DFrame) frame).featuresHelp();
			return true;
		case '^':
			((Life2DFrame) frame).referencesHelp();
			return true;
		case 'a':
		case 'A':
			((Life2DFrame) frame).aboutHelp();
			return true;
		case 'g':
			getLife();
			break;
		case 'w':
			writeLife();
			break;
		case 'G':
			getLifeFile();
			break;
		case 'W':
			writeLifeFile();
			break;
		case 'p':
		case 'P':
			printLife();
			break;
		case 'n':
		case 'N':
			System.out.println(getGridSignature());
			break;
		case 'm':
		case 'M':
			System.out.println(getGridStamp());
			break;
		case 't':
		case 'T':
			setSoupSearch(true, false);
			setPause(false);
			findLife(gliderSearch);
			break;
		case 'l':
		case 'L':
			setSoupSearch(false, true);
			setPause(false);
			findLife(gliderSearch);
			break;
		case 'c':
		case 'C':
			setPause(true);
			try {
				Thread.sleep(1);
			} catch (InterruptedException e) {
				//e.printStackTrace();
			}
			clearField(true);
			break;
		case ']':
		case '}':
			ncols++;
			((Life2DFrame) frame).setColumns(ncols);
			break;
		case '[':
		case '{':
			if (ncols > 4) {
				ncols--;
				((Life2DFrame) frame).setColumns(ncols);
			}
			break;
		case 'y':
		case 'Y':
			if (paused) {
				capturer.clearCapture();
				setSoupSearch(false, false);
				lifePulse(visible, false, true);
			} else {
				setPause(true);
				((Life2DFrame) frame).updateGen(gen);
				((Life2DFrame) frame).updateCells(nCells);
			}
			break;
		case '#':
			grid = !grid;
			drawGrid();
			break;
		case 'z':
		case 'Z':
			randomOrient = !randomOrient;
			break;
		case '|':
		case '\\':
		case 'r':
		case 'R':
			capturer.clearCapture();
			setSoupSearch(false, false);
			((Life2DFrame) frame).updateGen(gen);
			((Life2DFrame) frame).updateCells(nCells);
			setPause(!paused);
			banner("");
			break;
		case 'o':
		case 'O':
			wrap = !wrap;
			break;
		case 'v':
		case 'V':
			visible = !visible;
			break;
		case 's':
		case 'S':
			((Life2DFrame) frame).pauseGame();
			break;
		case '&':
			((Life2DFrame) frame).clearGame();
			break;
		case '~':
			((Life2DFrame) frame).setupGame();
			break;
		case '%':
			((Life2DFrame) frame).searchParamGame();
			break;
		case '$':
			getSignatureList(true, false, false, null);
			break;
		case '@':
			generateDB(true);
			break;
		case '*':
			((Life2DFrame) frame).showBuffer();
			break;
		case 'q':
		case 'Q':
		case 'x':
		case 'X':
		case '\03': /* Control-C */
			quitGame();
			break;
		default:
			return false; 
		}
		return true;
	}

	void findFileLife(int shake) {
		if (!paused) {
			paused = true; // hard code so title does not flash
		}
		if (signatures == null)
			generateDB(false);
		clearField(true);
		getLifeString(this.file0String, 0, 0);
		int x = 0, y = 0;
		if (shake > 0) {
			//Random generator = new Random(System.nanoTime());
			//while (x != -shake && x != shake &&
			//		y != -shake && y != shake) {
			//	x = generator.nextInt(2 * shake + 1) - shake;
			//	y = generator.nextInt(2 * shake + 1) - shake;
			//}
			while (x != -shake && x != shake &&
					y != -shake && y != shake) {
				x = X - shake;
				y = Y - shake;
				X++;
				if (X > 2 * shake) {
					X = 0;
					Y++;
					if (Y > 2 * shake) {
						Y = 0;
					}
				}
			}
			System.out.println("offset: " + x + " " + y);
		}
		getLifeString(this.file1String, x, y);
		checkFind = -1;
		paused = false;
	}

	
	void findLife(boolean allowGliderPatterns) {
		if (!paused) {
			paused = true; // hard code so title does not flash
		}
		if (signatures == null)
			generateDB(false);
		clearField(true);
		// some symmetries are counterproductive for gliders
		int symmetry;
		if (numberedSearch != -1)
			symmetry = numberedSearch;
		else
			symmetry = soup.randomSymmetry(getPolygon(), allowGliderPatterns);
		soup.displayRandomSoup(symmetry, fillPercent, stirs,
			Life2DSoup.soupSize(ncols),
			Life2DSoup.soupSize(nrows));
		checkFind = -1;
		paused = false;
	}

	void redrawAll() {
		repaint();
	}

	@Override
	public void mousePressed(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if (mouseDown)
			return;

		Point p = cellPl.screenToCell(event.getX(), event.getY());

		if (p == null)
			return;
		mouseDown = true;
		if ((modifiers & InputEvent.CTRL_DOWN_MASK) == 0) {
			if ((modifiers & InputEvent.BUTTON1_DOWN_MASK) != 0) {
				capturer.clearCapture();
				if (p.x >= 0 && p.x < ncols &&
						p.y >= 0 && p.y < nrows) {
					cellState = incCell(p.x, p.y);
				}
			}
		} else {
			capturer.clearCapture();
			capturer.startCapture(p.x, p.y);
		}
	}

	@Override
	public void mouseClicked(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	@Override
	public void mouseReleased(MouseEvent event) {
		if (mouseDown) { // Do not care which mouse button
			mouseDown = false;
		}
		capturer.stopCapture();
	}

	@Override
	public void mouseEntered(MouseEvent event) {
		focus = true;
		if (!scroll)
			drawFrame(focus);
		//banner(null);
	}

	@Override
	public void mouseExited(MouseEvent event) {
		focus = false;
		if (!scroll)
			drawFrame(focus);
	}

	@Override
	public void mouseDragged(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if ((modifiers & InputEvent.BUTTON3_DOWN_MASK) == 0) {
			Point p = cellPl.screenToCell(event.getX(), event.getY());

			if (p == null)
				return;
			mouseDown = true;
			if ((modifiers & InputEvent.CTRL_DOWN_MASK) == 0) {
				capturer.clearCapture();
				if (p.x >= 0 && p.x < ncols &&
						p.y >= 0 && p.y < nrows) {
					setCell(p.x, p.y, cellState, true);
				}
			} else {
				capturer.continueCapture(p.x, p.y);
			}
		}
	}

	@Override
	public void mouseMoved(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	public final void start() {
		if (gameRunner == null) {
			gameRunner = new Thread(this);
			gameRunner.start();
		}
	}

	public void stop() {
		/*if (gameRunner.isAlive())
			gameRunner.stop();*/
		gameRunner = null;
	}

	@Override
	public void run() {
		inits();
		/*sizeInit();*/
		repaint();
		while (!quit) {
			try {
				if (!generatingDB) {
					processAdvance();
					if (Thread.interrupted()) {
						throw new InterruptedException();
					}
					int delay = 20 * (Life2DInterface.MAX_SPEED +
						Life2DInterface.MIN_SPEED - speed) - 19;

					if (delay > 0)
						Thread.sleep(delay);
				} else {
					Thread.sleep(1000);
					generatingDB = false;
				}
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		if ((Life2DFrame) frame != null)
			((Life2DFrame) frame).quit();
	}

	@Override
	public void keyPressed(KeyEvent event) {
		if (event.getKeyCode() == KeyEvent.VK_ESCAPE) {
			((Life2DFrame) frame).shuffleDown();
			return;
		}
		processKey(event.getKeyChar());
	}

	@Override
	public void keyReleased(KeyEvent event) {
		/* Method required but keyPressed does work */
	}

	@Override
	public void keyTyped(KeyEvent event) {
		/* Method required but keyPressed does work */
	}

	public boolean getPause() {
		return paused;
	}

	public void setPause(boolean paused) {
		if (paused == this.paused)
			return;
		this.paused = paused;
		if (paused)
			banner(Life2DInterface.MSG_PAUSED);
		else
			banner("");
	}

	public void setSoupSearch(boolean patternSearch, boolean gliderSearch) {
		this.patternSearch = patternSearch;
		this.gliderSearch = gliderSearch;
	}

	public void setSoupSearch(int numberedSearch) {
		this.numberedSearch = numberedSearch;
		this.patternSearch = true;
	}

	public void setFilesSearch(String file0, String file1, int shake) {
		StringBuilder string0 = new StringBuilder();
		getLifeFile(file0, string0);
		StringBuilder string1 = new StringBuilder();
		getLifeFile(file1, string1);
		this.file0String = string0.toString();
		this.file1String = string1.toString();
		this.fileSearchShake = shake;
	}

	public synchronized void setDelay(int delay) {
		this.speed = Life2DInterface.MAX_SPEED +
			Life2DInterface.MIN_SPEED - 1 - delay;
	}

	public void setStates(int states) {
		this.states = states;
	}

	public void setNeighborIndex(int neighborIndex) {
		neighbors = Life2DRules.neighborToIndex(neighborIndex);
		if (grid)
			drawGrid();
	}

	public void setRandomOrient(boolean flag) {
		this.randomOrient = flag;
	}

	public void setGrid(boolean flag) {
		this.grid = flag;
		drawGrid();
	}

	public void setWrap(boolean flag) {
		this.wrap = flag;
	}

	public void setCellVisible(boolean flag) {
		this.visible = flag;
	}

	public void setMono(boolean flag) {
		this.mono = flag;
	}

	public void setRound(boolean flag) {
		this.round = flag;
		eraseFrame();
		drawField(true);
	}

	public void setFindStable(boolean flag) {
		this.findStable = flag;
	}

	public void setUnboundDetection(boolean flag) {
		this.unboundDetection = flag;
	}

	public void setEdgeDetection(boolean flag) {
		this.edgeDetection = flag;
	}

	public void setColumns(int columns) {
		if (this.ncols != getRealColumns(columns) ||
			this.nrows != getRowsFromColumns(columns)) {
			eraseFrame();
			this.ncols = getRealColumns(columns);
			this.nrows = getRowsFromColumns(columns);
			/*sizeInit();*/
			resizeField();
			drawField(true);
		}
	}

	public void setExcludePeriod(int period) {
		this.excludePeriod = period;
	}

	public int getExcludePeriod() {
		return this.excludePeriod;
	}

	public void setExcludeCount(int count) {
		this.excludeCount = count;
	}

	public int getExcludeCount() {
		return this.excludeCount;
	}

	public void setSearchCount(int count) {
		this.searchCount = count;
	}

	public int getSearchCount() {
		return this.searchCount;
	}

	public void setSearchLength(int length) {
		this.searchLength = length;
	}

	public int getSearchLength() {
		return this.searchLength;
	}

	public void setFillPercent(int percent) {
		this.fillPercent = percent;
	}

	public int getFillPercent() {
		return this.fillPercent;
	}

	public void setBlobPercent(int percent) {
		this.blobPercent = percent;
	}

	public int getBlobPercent() {
		return this.blobPercent;
	}

	public void setStirs(int count) {
		this.stirs = count;
	}

	public int getStirs() {
		return this.stirs;
	}

	public void setNumberedSearch(int count) {
		this.numberedSearch = count;
	}

	public int getNumberedSearch() {
		return this.numberedSearch;
	}

	public void setFileSearch() {
		this.fileSearch = true;
	}

	public boolean getFileSearch() {
		return this.fileSearch;
	}

	public int getBirth() {
		return birth;
	}

	public int getSurvival() {
		return survival;
	}

	public static void main(String[] args) {
		ArgumentParser argumentParser = new ArgumentParser(args);
		int speed = Life2DInterface.MAX_SPEED +
			Life2DInterface.MIN_SPEED - 1;
		int columns = ArgumentParser.parseOption(argumentParser.getOption("columns"),
			Life2DInterface.DEFAULT_COLUMNS);
		boolean randomOrient = true;
		boolean grid = false;
		boolean wrap = (ArgumentParser.parseOption(argumentParser.getOption("wrap"), 1) != 0); // TODO
		boolean visible = false;
		boolean mono = false;
		boolean round = false;
		boolean scroll = false;
		new Life2DRules(Life2DInterface.XML_FILE, null);
		new Life2DForms(Life2DInterface.XML_FILE, null);

		int states = ArgumentParser.parseOption(argumentParser.getOption("states"), 2);
		int neighbors = ArgumentParser.parseOption(argumentParser.getOption("neighbors"),
			Life2DRules.indexToNeighbor[Life2DInterface.DEFAULT_NEIGHBOR_INDEX]);

		String ruleString = argumentParser.getOption("rule");
		Life2D life2d = new Life2D(null, speed, columns,
			states, neighbors,
			randomOrient, grid,
			wrap, visible, mono,
			round, scroll, null);
		int searchCount = ArgumentParser.parseOption(argumentParser.getOption("searchCount"),
			Life2DInterface.DEFAULT_SEARCH_COUNT);
		life2d.setSearchCount(searchCount);
		int excludePeriod = ArgumentParser.parseOption(argumentParser.getOption("excludePeriod"),
			Life2DInterface.DEFAULT_EXCLUDE_PERIOD);
		life2d.setExcludePeriod(excludePeriod);
		int stirs = ArgumentParser.parseOption(argumentParser.getOption("stirs"),
			Life2DInterface.DEFAULT_STIRS);
		life2d.setStirs(stirs);
		int searchLength = ArgumentParser.parseOption(argumentParser.getOption("searchLength"),
			Life2DInterface.DEFAULT_SEARCH_LENGTH);
		life2d.setSearchLength(searchLength);
		boolean patternSearch = (ArgumentParser.parseOption(argumentParser.getOption("patternSearch"), 0) != 0);
		boolean gliderSearch = (ArgumentParser.parseOption(argumentParser.getOption("gliderSearch"), 0) != 0);
		life2d.setSoupSearch(patternSearch, gliderSearch);
		boolean edgeDetection = (ArgumentParser.parseOption(argumentParser.getOption("edgeDetection"), 0) != 0);
		life2d.setEdgeDetection(edgeDetection);
		boolean unboundDetection = (ArgumentParser.parseOption(argumentParser.getOption("unboundDetection"), 0) != 0);
		life2d.setUnboundDetection(unboundDetection);
		int blobPercent = ArgumentParser.parseOption(argumentParser.getOption("blobPercent"),
			Life2DInterface.DEFAULT_BLOB_PERCENT);
		life2d.setBlobPercent(blobPercent);
		int fillPercent = ArgumentParser.parseOption(argumentParser.getOption("fillPercent"),
			Life2DInterface.DEFAULT_FILL_PERCENT);
		life2d.setFillPercent(fillPercent);
		int numberedSearch = ArgumentParser.parseOption(argumentParser.getOption("numberedSearch"),
			-1);
		life2d.setNumberedSearch(numberedSearch);

		//String newRule = Life2DRules.parse(neighbors + ruleString);
		//life2d.param.setNeighborIndex(Life2DRules.neighborToIndex(neighbors));
		//life2d.param.setRuleName(newRule);
		life2d.param = new Life2DRules(neighbors, ruleString);
		life2d.signatures = null;
		life2d.generateDB(false);
		life2d.clearField(true);
		life2d.paused = false;
		life2d.start();
	}
}
