package com.sillycycle.bagleyd.life2d;

/*
 * @(#)Life2DRules.java
 *
 * Copyright 2008 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * The <code>Life2DRules</code> class handles the rules of life.
 * The way implemented should be good for 2 or 3 dimensional life.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Life2DRules {
	static final int MAX_NEIGHBORS = 12; // 12 < 32 bits
	static final int MAX_NEIGHBOR_INDEXES = 10;
	static final int MAX_GROUPS = 7; // More for triangularNT
	static final int FIRST_GROUP = 1; /* For Moore, corner or edge so group can start with 1 */
	static final int BASE = 10;
	static final boolean DEBUG = false;

	private Random generator;
	private String commonName;
	private String discoverer;
	private int neighborIndex;
	private String ruleName;
	private int state = 2, birth, survival;
	private int birthStir, survivalStir;

	private int[] birthGroup = new int[MAX_GROUPS];
	private int[] survivalGroup = new int[MAX_GROUPS];

	static byte[] maxgroups = {0, 1, 0, 4, 0, 0, 7, 0, 0, 0};
	static byte[][] groupNumber = {
		{0, 0,  0,  0,  0, 0, 0},	/* 3 Triangular, would not make sense to use */
		{0, 0,  0,  0,  0, 0, 0},	/* 4 MAX_GROUP 2 e/i future improvement */
		{0, 0,  0,  0,  0, 0, 0},	/* 5 */
		{0, 3,  3,  3,  0, 0, 0},	/* 6 Hexagonal */
		{0, 0,  0,  0,  0, 0, 0},	/* alt 6 */
		{0, 0,  0,  0,  0, 0, 0},	/* 7 */
		{2, 6, 10, 14, 10, 6, 2},	/* 8 Moore */
		{0, 0,  0,  0,  0, 0, 0},	/* 9 Future improvement */
		{0, 0,  0,  0,  0, 0, 0},	/* alt 9 */
		{3, 0,  0,  0,  0, 0, 0}	/* 12 Future improvement, would need more values */
	};

	// conwaylife.com/wiki/Isotropic_non-totalistic_Life-like_cellular_automaton
	// Added 'f' as appears not all orientations represented

	// Group stuff hack for Andreen and Callahan Hexagonal rules
	// B2o/S2b34: Birth of x if 2o,
	// Survival of x if 2m, 3, or 4 neighbors
	//       O O            O .            O .
	// 2o:  . x .     2m:  . x O     2p:  . x .
	//       . .            . .            . O
        // Also Bob Andreen's rule
	// B2o3o4m/S2om4o (original notation: 234'B/22'4S)
	//       O O            O O            O .
	// 3o:  . x O     3m:  . x .     3p:  . x O
	//       . .            . O            O .
	//
	//       O O            O O            O O
	// 4o:  . x O     4m:  . x O     4p:  . x .
	//       . O            O .            O O
	// Some other rules
	// B2o3mp5/S12mp3o4mp6
	// B23op4mp6/S12mp3p4o56
	// B2op6/S13m
	static char[] nontotalistic6 = // 2^6 values
	{'x', 'x', 'x', 'o', 'x', 'm', 'o', 'o',
	'x', 'p', 'm', 'm', 'o', 'm', 'o', 'o',
	'x', 'm', 'p', 'm', 'm', 'p', 'm', 'm',
	'o', 'm', 'm', 'p', 'o', 'm', 'o', 'x',
	'x', 'o', 'm', 'o', 'p', 'm', 'm', 'o',
	'm', 'm', 'p', 'm', 'm', 'p', 'm', 'x',
	'o', 'o', 'm', 'o', 'm', 'm', 'p', 'x',
	'o', 'o', 'm', 'x', 'o', 'x', 'x', 'x'};

	// Maydwell 8B3/S2ckan3 (original notation 3/2ab3) (no 2 edges)
	// 'f' seems to be "forgotten" in life wiki
	//     ..*      .*.
	// 1c: .x.  1e: .x.
	//     ...      ...
	//
	//     ..*      .*.      .*.      .**      .*.      ..*
	// 2c: .x.  2e: .x*  2k: .x.  2a: .x.  2i: .x.  2n: .x.
	//     ..*      ...      ..*      ...      .*.      *..
	//
	//     ..*      .*.      .*.      .**      ..*      ..*
	// 3c: .x.  3e: .x*  3k: *x.  3a: .x*  3i: .x*  3n: .x.
	//     *.*      .*.      ..*      ...      ..*      .**
	//
	//     ..*      ..*      ..*      ..*
	// 3y: *x.  3q: .x*  3j: .x*  3r: *x*
	//     ..*      *..      .*.      ...
	//
	//     *.*      .*.      .**      .**      .**      ..*
	// 4c: .x.  4e: *x*  4k: *x.  4a: .x*  4i: .x.  4n: .x.
	//     *.*      .*.      ..*      ..*      .**      ***
	//
	//     ..*      .**      ..*      .**      ..*      **.      ..*      **.
	// 4y: *x.  4q: .x*  4j: *x*  4r: *x*  4t: *x*  4w: .x*  4z: *x*  4f: .x*
	//     *.*      *..      .*.      ...      ..*      ..*      *..      *..
	//
	//     **.      *.*      *.*      *..      **.      **.
	// 5c: *x*  5e: *x.  5k: .x*  5a: *x.  5i: *x.  5n: *x*
	//     .*.      *.*      **.      ***      **.      *..
	//
	//     **.      **.      **.      **.
	// 5y: .x*  5q: *x.  5j: *x.  5r: .x.
	//     **.      .**      *.*      ***
	//
	//     **.      *.*      *.*      *..      *.*      **.
	// 6c: *x*  6e: *x.  6k: *x*  6a: *x*  6i: *x*  6n: *x*
	//     **.      ***      **.      ***      *.*      .**
	//
	//     **.      *.*
	// 7c: *x*  7e: *x*
	//     ***      ***

	static char[] nontotalistic8 = // 2^8 values
	//0    1    2    3    4    5    6    7    8    9    a    b    c    d    e    f
	{'x', 'c', 'e', 'a', 'c', 'c', 'a', 'i', 'e', 'k', 'e', 'k', 'a', 'n', 'a', 'a',
	'c', 'n', 'k', 'q', 'c', 'c', 'n', 'n', 'a', 'q', 'k', 'w', 'i', 'n', 'a', 'a',
	'e', 'k', 'i', 'r', 'k', 'y', 'r', 't', 'e', 'k', 'e', 'j', 'k', 'f', 'r', 'n',
	'a', 'q', 'r', 'z', 'n', 'y', 'i', 'r', 'a', 'q', 'r', 'q', 'a', 'j', 'i', 'a',
	'c', 'c', 'k', 'n', 'n', 'c', 'q', 'n', 'k', 'y', 'k', 'f', 'q', 'y', 'q', 'j',
	'c', 'c', 'y', 'y', 'c', 'c', 'y', 'e', 'n', 'y', 'f', 'k', 'n', 'e', 'j', 'e',
	'a', 'n', 'r', 'i', 'q', 'y', 'z', 'r', 'k', 'f', 'j', 'y', 'w', 'k', 'q', 'k',
	'i', 'n', 't', 'r', 'n', 'e', 'r', 'i', 'a', 'j', 'n', 'k', 'a', 'e', 'a', 'e',
	'e', 'a', 'e', 'a', 'k', 'n', 'k', 'a', 'i', 'r', 'e', 'r', 'r', 'i', 'r', 'i',
	'k', 'q', 'k', 'q', 'y', 'y', 'f', 'j', 'r', 'z', 'j', 'q', 't', 'r', 'n', 'a',
	'e', 'k', 'e', 'r', 'k', 'f', 'j', 'n', 'e', 'j', 'e', 'c', 'j', 'y', 'c', 'c',
	'k', 'w', 'j', 'q', 'f', 'k', 'y', 'k', 'r', 'q', 'c', 'n', 'n', 'k', 'c', 'c',
	'a', 'i', 'k', 'a', 'q', 'n', 'w', 'a', 'r', 't', 'j', 'n', 'z', 'r', 'q', 'a',
	'n', 'n', 'f', 'j', 'y', 'e', 'k', 'e', 'i', 'r', 'y', 'k', 'r', 'i', 'k', 'e',
	'a', 'a', 'r', 'i', 'q', 'j', 'q', 'a', 'r', 'n', 'c', 'c', 'q', 'k', 'n', 'c',
	'a', 'a', 'n', 'a', 'j', 'e', 'k', 'e', 'i', 'a', 'c', 'c', 'a', 'e', 'c', 'x'};

	// letters chosen may change...
	//	
	// 1c *     1e .     1k .
	//   . .      . *      . .
	//  . x .    . x .    . x *
	//  . . .    . . .    . . .
	// . . . .  . . . .  . . . .

	static char[] nontotalistic12 =
	{'x', 'c', 'e', 'x', 'k', 'x', 'x', 'x', 'e', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'c', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'e', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'k', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'e', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'c', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'e', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'k', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'e', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x',
	'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'};
	
	public Life2DRules(String fileName, URL url) {
		readRulesXML(fileName, url, true);
		readRulesXML(fileName, url, false);
		generator = new Random(System.nanoTime());
	}

	static void readRulesXML(String fileName, URL url, boolean pre) {
		try {
			// Convert to JSON to Disable XML external entity (XXE) processing.
			DocumentBuilderFactory docBuilderFactory =
				DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder =
				docBuilderFactory.newDocumentBuilder();
			Document doc;

			if (url == null) {
				doc = docBuilder.parse(new File(fileName));
			} else {
				doc = docBuilder.parse(new URL(url, fileName).openStream());
				//System.out.println("Loaded xml: " + url + fileName);
			}
			int[] ruleIndex = new int[Life2DRules.MAX_NEIGHBOR_INDEXES];
			int ni, ri;
			// normalize text representation
			doc.getDocumentElement().normalize();
			if (DEBUG)
				System.out.println("Root element of the doc is " +
					doc.getDocumentElement().getNodeName());

			NodeList listOfRules =
				doc.getElementsByTagName("rule");
			if (DEBUG)
				System.out.println("Total number of rules: " + listOfRules.getLength());
			if (pre)
				lifeRules = new Life2DRules[Life2DRules.MAX_NEIGHBOR_INDEXES][];
			for (int r = 0; r < listOfRules.getLength(); r++) {
				String neighborRuleString;

				Node ruleNode = listOfRules.item(r);
				if (ruleNode.getNodeType() != Node.ELEMENT_NODE) {
					continue;
				}
				Element ruleElement = (Element) ruleNode;

				NodeList rulenameList = ruleElement.getElementsByTagName("rulename");
				Element rulenameElement = (Element)rulenameList.item(0);

				NodeList textRuleNameList = rulenameElement.getChildNodes();
				neighborRuleString = (textRuleNameList.item(0)).getNodeValue().trim();
				ni = Life2DRules.parseNeighborFromRule(neighborRuleString);
				ri = ruleIndex[ni];
				ruleIndex[ni]++;
				if (pre) {
					continue;
				}
				String commonname = "", discoverer = "";
				NodeList commonnameList = ruleElement.getElementsByTagName("commonname");
				Element commonnameElement = (Element)commonnameList.item(0);

				NodeList textCommonNameList = commonnameElement.getChildNodes();
				if ((textCommonNameList.item(0)) != null)
					commonname = (textCommonNameList.item(0)).getNodeValue().trim();

				NodeList discovererList = ruleElement.getElementsByTagName("discoverer");
				Element discovererElement = (Element)discovererList.item(0);

				NodeList textDiscovererList = discovererElement.getChildNodes();
				if ((textDiscovererList.item(0)) != null)
					discoverer = (textDiscovererList.item(0)).getNodeValue().trim();

				if (DEBUG)
					System.out.println(commonname + " " + discoverer + " " +
						Life2DRules.parseNeighborRule(neighborRuleString));
				lifeRules[ni][ri] = new Life2DRules(commonname, discoverer,
					Life2DRules.parseNeighborRule(neighborRuleString));
			}
			if (pre) {
				for (ni = 0; ni < Life2DRules.MAX_NEIGHBOR_INDEXES; ni++) {
					if (DEBUG)
						System.out.println("ni: " + ni + ", " + ruleIndex[ni]);
					lifeRules[ni] = new Life2DRules[ruleIndex[ni]];
				}
			}
		} catch (SAXParseException err) {
			System.out.println("** Parsing error" + ", line " +
				err.getLineNumber() + ", uri " +
				err.getSystemId());
			System.out.println(" " + err.getMessage());

		} catch (SAXException e) {
			Exception x = e.getException();
			((x == null) ? e : x).printStackTrace();

		} catch (ParserConfigurationException | IOException | DOMException e) {
			e.printStackTrace();
		}
	}

	public Life2DRules(String ruleName, int birth, int survival) {
		this.commonName = "";
		this.discoverer = "";
		this.ruleName = ruleName;
		generator = new Random(System.nanoTime());
		this.birth = birth;
		this.survival = survival;
		this.birthStir = birth;
		this.survivalStir = survival;
	}

	public Life2DRules(String commonName, String discoverer,
			String ruleName) {
		this.commonName = commonName;
		this.discoverer = discoverer;
		this.ruleName = ruleName;
		generator = new Random(System.nanoTime());
		if (ruleName != null && !ruleName.isEmpty()) {
			setMasks(getBirthString(ruleName),
				getSurvivalString(ruleName));
		}
	}

	public Life2DRules(int neighbors, String ruleName) {
		this.commonName = "";
		this.discoverer = "";
		this.neighborIndex = neighborToIndex(neighbors);
		this.ruleName = ruleName;
		generator = new Random(System.nanoTime());
		if (ruleName != null && !ruleName.isEmpty()) {
			setMasks(getBirthString(ruleName),
				getSurvivalString(ruleName));
		}
		init();
	}

	public Life2DRules(int neighbors, int birth, int survival,
			int[] birthGroup, int[] survivalGroup) {
		StringBuilder birthString = new StringBuilder("B");
		StringBuilder survivalString = new StringBuilder("S");
		this.neighborIndex = neighborToIndex(neighbors);
		generator = new Random(System.nanoTime());
		this.birth = birth;
		this.survival = survival;
		this.birthStir = birth;
		this.survivalStir = survival;

		this.birthGroup = birthGroup;
		this.survivalGroup = survivalGroup;
		for (int i = 0; i <= MAX_NEIGHBORS; i ++) {
			int mask = 1 << i;
			boolean found = false;
			int bPlaced = -1;
			int sPlaced = -1;
			//System.out.println(birth + " birthGroup=" + toStringGroup(birthGroup));
			//System.out.println(survival + " survivalGroup=" + toStringGroup(survivalGroup));

			if (i >= FIRST_GROUP && i < FIRST_GROUP + maxgroups[neighborIndex]) {
				for (int j = 0; j < groupNumber[neighborIndex][i - FIRST_GROUP]; j++) {
					//System.out.println("groupNumber " + groupNumber[neighborIndex][i - FIRST_GROUP]);
					if ((birthGroup[i - FIRST_GROUP] & (1 << j)) != 0) {
						found = true;
						if (bPlaced != i) {
							birthString.append(Integer.toString(i));
							bPlaced = i;
						}
						birthString.append(indexToNontot(neighbors, j));
						bPlaced = i;
						//System.out.println("birthString=" + birthString);
					}
					if ((survivalGroup[i - FIRST_GROUP] & (1 << j)) != 0) {
						found = true;
						if (sPlaced != i) {
							survivalString.append(Integer.toString(i));
							sPlaced = i;
						}
						survivalString.append(indexToNontot(neighbors, j));
						//System.out.println("survivalString=" + survivalString);
					}
				}
			}
			if (!found) {
				if ((mask & birth) != 0) {
					if (i >= BASE)
						birthString.append("_");
					birthString.append(i);
				}
				if ((mask & survival) != 0) {
					if (i >= BASE)
						survivalString.append("_");
					survivalString.append(i);
				}
			}
		}
		ruleName = birthString.toString() + "/" + survivalString.toString();
		commonName = "";
		for (Life2DRules lifeRule : lifeRules[neighborIndex]) {
			if (birth == lifeRule.birth && survival == lifeRule.survival) {
				commonName = lifeRule.commonName;
				break;
			}
		}
		discoverer = "";
		for (Life2DRules lifeRule : lifeRules[neighborIndex]) {
			if (birth == lifeRule.birth && survival == lifeRule.survival) {
				discoverer = lifeRule.discoverer;
				break;
			}
		}
		init();
	}

	protected static String toString(int[] values) {
		StringBuilder string = new StringBuilder();

		for (int i = 0; i < values.length; i++) {
			if (i != 0)
				string.append(", ");
			string.append(i).append(": ").append(values[i]);
		}
		return string.toString();
	}

	public String getCommonName() {
		return commonName;
	}

	public String getDiscoverer() {
		return discoverer;
	}

	public int getNeighbors() {
		return indexToNeighbor[neighborIndex];
	}

	public String getRuleName() {
		return ruleName;
	}

	public static String getBirthString(String ruleName) {
		try {
			String[] ruleParts = ruleName.split("/");

			return ruleParts[0].substring(ruleParts[0].indexOf('B') + 1);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public static int getBirth(String ruleName) {
		return setMask(full(getBirthString(ruleName)));
	}

	public static int[] getBirthGroup(int neighborIndex, String ruleName) {
		String[] ruleParts = ruleName.split("/");

		return setMaskGroup(neighborIndex, ruleParts[0].substring(ruleParts[0].indexOf('B') + 1));
	}

	public static String getSurvivalString(String ruleName) {
		try {
			String[] ruleParts = ruleName.split("/");

			if (ruleParts.length > 1)
				return ruleParts[1].substring(ruleParts[1].indexOf('S') + 1);
			System.out.println("Error" + ruleName);
			return null;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public static int getSurvival(String ruleName) {
		return setMask(full(getSurvivalString(ruleName)));
	}

	public static int[] getSurvivalGroup(int neighborIndex, String ruleName) {
		String[] ruleParts = ruleName.split("/");

		return setMaskGroup(neighborIndex, ruleParts[1].substring(ruleParts[1].indexOf('S') + 1));
	}

	public static String getRuleName(int neighborIndexValue, int ruleIndexValue) {
		return lifeRules[neighborIndexValue][ruleIndexValue].ruleName;
	}

	public int getState() {
		return state;
	}

	public int getBirth() {
		return birth;
	}

	public int getSurvival() {
		return survival;
	}

	public int[] getBirthGroup() {
		return birthGroup;
	}

	public int[] getSurvivalGroup() {
		return survivalGroup;
	}

	public void setCommonName(String commonName) {
		this.commonName = commonName;
	}

	public void setDiscoverer(String discoverer) {
		this.discoverer = discoverer;
	}

	public void setNeighborIndex(int neighborIndex) {
		this.neighborIndex = neighborIndex;
	}

	public void setNeighbors(int neighbors) {
		this.neighborIndex = neighborToIndex(neighbors);
	}

	public void setRuleName(String ruleName) {
		this.ruleName = ruleName;
	}

	public void setState(int state) {
		this.state = state;
	}

	public void setBirth(int birth) {
		this.birth = birth;
	}

	public void setSurvival(int survival) {
		this.survival = survival;
	}

	public boolean getBirth(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return ((this.birth & (1 << count)) != 0);
		}
		return false;
	}

	public boolean getSurvival(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return ((this.survival & (1 << count)) != 0);
		}
		return false;
	}

	public boolean getBirth(int count, int group) {
		int neighbors = indexToNeighbor[neighborIndex];

		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return (getBirth(count) ||
				(count >= FIRST_GROUP &&
				count < FIRST_GROUP + maxgroups[neighborIndex] &&
				((neighbors == 6 &&
				(birthGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic6[group]))) != 0) ||
				(neighbors == 8 &&
				(birthGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic8[group]))) != 0) ||
				(neighbors == 12 &&
				(birthGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic12[group]))) != 0))));
		}
		return false;
	}

	public boolean getSurvival(int count, int group) {
		int neighbors = indexToNeighbor[neighborIndex];

		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return (getSurvival(count) ||
				(count >= FIRST_GROUP &&
				count < FIRST_GROUP + maxgroups[neighborIndex] &&
				((neighbors == 6 &&
				(survivalGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic6[group]))) != 0) ||
				(neighbors == 8 &&
				(survivalGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic8[group]))) != 0) ||
				(neighbors == 12 &&
				(survivalGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic12[group]))) != 0))));
		}
		return false;
	}

	public boolean getBirthStir(int count, int group) {
		int neighbors = indexToNeighbor[neighborIndex];

		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return (getBirthStir(count) ||
				(count >= FIRST_GROUP &&
				count < FIRST_GROUP + maxgroups[neighborIndex] &&
				((neighbors == 6 &&
				(birthGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic6[group]))) != 0) ||
				(neighbors == 8 &&
				(birthGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic8[group]))) != 0) ||
				(neighbors == 12 &&
				(birthGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic12[group]))) != 0))));
		}
		return false;
	}

	public boolean getSurvivalStir(int count, int group) {
		int neighbors = indexToNeighbor[neighborIndex];

		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return (getSurvivalStir(count) ||
				(count >= FIRST_GROUP &&
				count < FIRST_GROUP + maxgroups[neighborIndex] &&
				((neighbors == 6 &&
				(survivalGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic6[group]))) != 0) ||
				(neighbors == 8 &&
				(survivalGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic8[group]))) != 0) ||
				(neighbors == 12 &&
				(survivalGroup[count - FIRST_GROUP] &
				(1 << nontotToIndex(nontotalistic12[group]))) != 0))));
		}
		return false;
	}

	public boolean getBirthStir(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return ((this.birthStir & (1 << count)) != 0);
		}
		return false;
	}

	public boolean getSurvivalStir(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return ((this.survivalStir & (1 << count)) != 0);
		}
		return false;
	}

	// Use dash for 3 or more consecutive number
	// This may be problematic as "-" may mean all but... for nontotalistic
	public static String condense(String string) {
		StringBuilder sb = new StringBuilder();
		int previous = -2, start = -2;
		int digits = 0;
		boolean gotUnderscore = false;
		boolean putDash = false;
		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);

			if (c == '_') {
				gotUnderscore = true;
				digits = 0;
			} else if (c >= '0' && c <= '9') {
				if (gotUnderscore) {
					digits = (c - '0') * BASE;
					gotUnderscore = false;
				} else {
					int current = digits + (c - '0');
					digits = 0;
					if (current == previous + 1) {
						if (!putDash) {
							if (previous >= BASE) {
								sb.append("_");
							}
							sb.append(previous);
							start = previous;
							putDash = true;
						}
						previous = current;
					} else {
						if (previous >= 0) {
							if (putDash && previous != start + 1) {
								sb.append("-");
							}
							if (previous >= BASE) {
								sb.append("_");
							}
							sb.append(previous);
						}
						previous = current;
						putDash = false;
						start = -2;
					}
				}
			} else {
				if (previous >= 0) {
					if (previous >= BASE) {
						sb.append("_");
					}
					sb.append(previous);
				}
				sb.append(c);
				gotUnderscore = false;
				putDash = false;
				digits = 0;
				previous = -2;
				start = -2;
			}
		}
		if (previous >= 0) {
			if (putDash && previous != start + 1) {
				sb.append("-");
			}
			if (previous >= BASE) {
				sb.append("_");
			}
			sb.append(previous);
		}
		return sb.toString();
	}

	public static String full(String string) {
		StringBuilder sb = new StringBuilder();
		int previous = -2;
		int digits = 0;
		boolean gotUnderscore = false;
		boolean gotDash = false;
		int current = -1;

		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);

			if (c == '-') {
				gotDash = true;
			} else if (c == '_') {
				gotUnderscore = true;
				digits = 0;
			} else if (c >= '0' && c <= '9') {
				if (gotUnderscore) {
					digits = (c - '0') * BASE;
					gotUnderscore = false;
				} else {
					current = digits + (c - '0');
					digits = 0;
					if (gotDash) {
						for (int j = previous; j <= current; j++) {
							if (j >= BASE) {
								sb.append("_");
							}
							sb.append(j);
						}
						gotDash = false;
						previous = -2;
					} else {
					if (previous >= 0) {
						if (previous >= BASE) {
							sb.append("_");
							}
							sb.append(previous);
						}
						previous = current;
					}
				}
			} else {
				if (previous >= 0) {
					if (previous >= BASE) {
						sb.append("_");
					}
					sb.append(previous);
				}
				if (gotDash) {
					sb.append('-');
				}
				sb.append(c);
				gotUnderscore = false;
				gotDash = false;
				digits = 0;
				previous = -2;
			}
		}
		if (previous >= 0) {
			if (previous >= BASE) {
				sb.append("_");
			}
			sb.append(previous);
		}
		return sb.toString();
	}

	public static int setMask(String string, int[] group) {
		int mask = 0, j = -1, extraDigits = 0, leadingDigit = 0, g;
		boolean first = true;
		for (int i = 0; i < MAX_GROUPS; i++)
			group[i] = 0;
		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);

			// Disable hexadecimal
			//if (c >= 'A' && c <= 'Z') {
			//	j = BASE + c - 'A';
			//	extraDigits = 0;
			//} else
			if (c >= 'a' && c <= 'z') {
				g = nontotToIndex(c);
				if (j != -1) {
					if (first)
						mask -= (1 << j);
					first = false;
					group[j - FIRST_GROUP] |= (1 << g);
					continue;
				}
			} else if (c >= '0' && c <= '9') {
				if (extraDigits == 0) {
					j = leadingDigit * BASE + (c - '0');
					leadingDigit = 0;
				} else {
					extraDigits--;
					leadingDigit = (c - '0');
					j = -1;
				}
				first = true;
			} else if (c == '_') {
				extraDigits = 1;
				j = -1;
				first = true;
			} else if (c == '-') {
				continue;
			} else {
				System.out.println("Extra: " + string.charAt(i));
				extraDigits = 0;
				j = -1;
				first = true;
			}
			if (j >= 0 && j <= MAX_NEIGHBORS)
				mask += (1 << j);
		}
		return mask;
	}

	public static int setMask(String string) {
		int mask = 0, j = -1, extraDigits = 0, leadingDigit = 0;
		boolean first = true;

		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);

			// Disable hexadecimal
			//if (c >= 'A' && c <= 'Z') {
			//	j = BASE + c - 'A';
			//	extraDigits = 0;
			//} else
			if (c >= 'a' && c <= 'z') {
				//g = nontotToIndex(c);
				if (j != -1) {
					if (first)
						mask -= (1 << j);
					first = false;
					//group[j - FIRST_GROUP] |= (1 << g);
					continue;
				}
			} else if (c >= '0' && c <= '9') {
				if (extraDigits == 0) {
					j = leadingDigit * BASE + (c - '0');
					leadingDigit = 0;
				} else {
					extraDigits--;
					leadingDigit = (c - '0');
					j = -1;
				}
				first = true;
			} else if (c == '_') {
				extraDigits = 1;
				j = -1;
				first = true;
			} else if (c == '-') {
				continue;
			} else {
				System.out.println("Extra: " + string.charAt(i));
				extraDigits = 0;
				j = -1;
				first = true;
			}
			if (j >= 0 && j <= MAX_NEIGHBORS)
				mask += (1 << j);
		}
		return mask;
	}

	public static int[] setMaskGroup(int neighborIndex, String string) {
		int j = -1, g;
		char c;
		int[] group = new int[MAX_GROUPS];
		boolean gotDash = false;
		for (int i = 0; i < MAX_GROUPS; i++)
			group[i] = 0;
		for (int i = 0; i < string.length(); i++) {
			c = string.charAt(i);
			if (c >= 'A' && c <= 'Z')
				j = BASE + c - 'A';
			else if (c >= 'a' && c <= 'z') {
				g = nontotToIndex(c);
				if (gotDash && j != -1) {
					ArrayList<Character> letters = new  ArrayList<>();
					letters.add(c);
					c = string.charAt(i);
					for (i++; i < string.length(); i++) {
						if (c >= 'a' && c <= 'z') {
							letters.add(c);
						} else {
							i--;
							break;
						}
					}
					ArrayList<Character> newLetters = invertGroupForChars(groupNumber[neighborIndex][j - 1], letters, indexToNeighbor[neighborIndex]);
					for (char letter: newLetters) {
						int newG = nontotToIndex(letter);
						group[j - FIRST_GROUP] |= (1 << newG); // just friends -rule=B2-a/S12  -4 bounds length 7
						gotDash = false;
					}
				} else {
					if (j != -1) {
						group[j - FIRST_GROUP] |= (1 << g); // just friends -rule=B2-a/S12  -4 bounds length 7
					}
				}
			} else if (c == '-')
				gotDash = true;
			else
				j = c - '0';
		}
		return group;
	}

	public void setBirthMask(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			this.birth |= (1 << count);
		}
		this.birthStir = this.birth;
	}

	public void setSurvivalMask(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			this.survival |= (1 << count);
		}
		this.survivalStir = this.survival;
	}

	public void resetBirthMask(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			this.birth &= ~(1 << count);
		}
		this.birthStir = this.birth;
	}

	public void resetSurvivalMask(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			this.survival &= ~(1 << count);
		}
		this.survivalStir = this.survival;
	}

	static final String[] defaultRuleStrings = {
		"B2/S12", "B23/S14", "B2/S24", "B2/S35", "B34/S23",
		"B346/S23", "B3/S23", "B34/S3", "B3/S34", "B4/S35"
	};

	static final Integer defaultRuleIndex(int ni) {
		if (lifeRules[ni] == null)
			return null;
		for (int ri = 0; ri < lifeRules[ni].length; ri++) {
			if (lifeRules[ni][ri] != null &&
					defaultRuleStrings[ni].equals(lifeRules[ni][ri].ruleName))
				return ri;
		}
		return null;
	}

	static Life2DRules[][] lifeRules;

	// Array of Generic Hash Tables
	ArrayList<HashMap<String, Integer>> arrayHashMap = new ArrayList<>();
//http://forums.sun.com/thread.jspa?threadID=530823
	final void init() {
		if (!arrayHashMap.isEmpty())
			arrayHashMap.clear();
		for (int i = 0; i < MAX_NEIGHBOR_INDEXES; i++) {
//System.out.println("Life2DRules.init: " + " " + i);
			arrayHashMap.add(new HashMap<String, Integer>());
//System.out.println("Life2DRules.init: " + " " + neighborIndex + " " + lifeRules[neighborIndex].length);
			for (int j = 0; j < lifeRules[i].length; j++) {
//System.out.println("Life2DRules.init: " + " " + indexToNeighbor[neighborIndex] + Life2DRules.parseRuleFull(lifeRules[i][j].ruleName) + " " + j);
				arrayHashMap.get(i).put(lifeRules[i][j].ruleName, j);
			}
		}
	}

	Integer getRuleIndex(int neighborhoodIndex, String s) {
		if (lifeRules[neighborhoodIndex].length == 0) {

			return null;
		}
//System.out.println("Life2DRules.getRuleIndex: ni " + neighborIndex + " " + lifeRules[neighborIndex].length);
//System.out.println("Life2DRules.getRuleIndex: value " + arrayHashMap.get(neighborIndex).values());
//System.out.println("Life2DRules.getRuleIndex: key " + arrayHashMap.get(neighborIndex).keySet());
//System.out.println("Life2DRules.getRuleIndex: n " + indexToNeighbor[neighborIndex] + " " + s + " " + arrayHashMap.get(neighborIndex).get(s));
		return arrayHashMap.get(neighborhoodIndex).get(s);
	}

	static String[] getStrings(int neighborIndex) {
		String[] strings = new String[lifeRules[neighborIndex].length + 1];
		int i;

		for (i = 0; i < lifeRules[neighborIndex].length; i++) {
			strings[i] = lifeRules[neighborIndex][i].ruleName;
			if (!"".equals(lifeRules[neighborIndex][i].commonName))

				strings[i] += ", " + lifeRules[neighborIndex][i].commonName;
			if (!"".equals(lifeRules[neighborIndex][i].discoverer))
				strings[i] += ", " + lifeRules[neighborIndex][i].discoverer;
		}
		strings[i] = "Other";
		return strings;
		//return (String[]) (arrayHashMap.get(neighborIndex).values()).toArray();
		//Collection<String> strings = arrayHashMap.get(neighborIndex).values();
		//return (String[]) strings.toArray();
		//String[] strings = new String[arrayHashMap.size()];
		//for (int i = 0; i < lifeRules[ni].length; i++) {
	//		strings[i] = arrayHashMap.get(neighborIndex).getKey(i);
		//}
		//return strings;
	}

	static final int[] neighborIndexToPolygon = {
		3, 4, 5, 6, 3, 5, 4, 3, 3, 3
	};

	static int neighborToPolygon(int neighbors) {
		switch (neighbors) {
		case 3: /* triangle */
		case -6:
		case 9:
		case -9:
		case 12:
			return 3;
		case 4: /* square */
		case 8:
			return 4;
		case 5: /* pentagon, Cairo Tiling */
		case 7:
			return 5;
		case 6: /* hexagon */
			return 6;
		default:
			System.out.println("no polygon known for " +
				neighbors + " neighbors");
			return 4;
		}
	}

	static final String[] neighborIndexStrings = {
		"3, Triangular, von Neuman",
		"4, Square, von Neuman",
		"5, Pentagonal, von Neuman, Cairo Tiling",
		"6, Hexagonal",
		"(-)6, Triangular, sides and far corners",
		"7, Pentagonal, Moore, Cairo Tiling",
		"8, Square, Moore",
		"9, Triangular, exclude far corners",
		"(-)9, Triangular, exclude sides",
		"12, Triangular, Moore"
	};

	static final int[] indexToNeighbor = {
		3, 4, 5, 6, -6, 7, 8, 9, -9, 12
	};

	static int neighborToIndex(int neighbors) {
		switch (neighbors) {
		case 3: /* von Neuman Neighbor Count */
			return 0;
		case 4: /* von Neuman Neighbor Count */
			return 1;
		case 5: /* Cairo Tiling, von Neuman Neighbor Count */
			return 2;
		case 6:
			return 3;
		case -6:
			return 4;
		case 7: /* Cairo Tiling, Moore Neighbor Count */
			return 5;
		case 8: /* Moore Neighbor Count */
			return 6;
		case 9: /* Moore Neighbor Count but excluding far corners */
			return 7;
		case -9: /* Moore Neighbor Count but excluding far corners */
			return 8;
		case 12: /* Moore Neighbor Count */
			return 9;
		default:
			System.out.println("no neighborhood like " +
				neighbors + " known");
			return Life2DInterface.DEFAULT_NEIGHBOR_INDEX;
		}
	}

	static char[] hexagonalNontot = {'o', 'm', 'p'};
	static char[] mooreNontot = {'c', 'e', 'k', 'a', 'i', 'n',
		'y', 'q', 'j', 'r', 't', 'w', 'z', 'f'};
	static char[] triangularNontot = {'c', 'e', 'k'}; // experimental

	static char indexToNontot(int neighbors, int i) {
		if (neighbors == 6)
			return hexagonalNontot[i];
		else if (neighbors == 8)
			return mooreNontot[i];
		else
			return triangularNontot[i];
	}

	static ArrayList<Character> invertGroupForChars(int group, ArrayList<Character> letters, int neighbors) {
		ArrayList<Character> arrayList = new ArrayList<>();
		for (int i = 0; i < group; i++) {
			boolean found = false;
			char newLetter = indexToNontot(neighbors, i);
			for (char letter: letters) {
				if (newLetter == letter) {
					found = true;
					break;
				}
			}
			if (!found) {
				arrayList.add(newLetter);
			}
		}
		return arrayList;
	}

	static int nontotToIndex(char letter) {
		switch (letter) { // take advantage that there is no overlap
		case 'c':
		case 'o':
			return 0;
		case 'e':
		case 'm':
			return 1;
		case 'k':
		case 'p':
			return 2;
		case 'a':
			return 3;
		case 'i':
			return 4;
		case 'n':
			return 5;
		case 'y':
			return 6;
		case 'q':
			return 7;
		case 'j':
			return 8;
		case 'r':
			return 9;
		case 't':
			return 10;
		case 'w':
			return 11;
		case 'z':
			return 12;
		case 'f':
			return 13;
		default: // 'x':
			return 14;
		}
	}


	static String parse(String rule) {
		return indexToNeighbor[parseNeighborFromRule(rule)] +
			parseNeighborRule(rule);
	}

	static int parseStateFromRule(String rule) {
		int neighbor = 0;
		int state = 2;
		int i = 0;
		boolean negative = false;

		for (; i < rule.length(); i++) {
			char letter = rule.charAt(i);

			if (letter >= '0' && letter <= '9')
				neighbor = neighbor * BASE + (letter - '0');
			else if (letter == '-')
				negative = !negative;
			else if (letter == 'N')
				// really reading state not neighbor
				return neighbor;
			else
				break;
		}
		return state;
	}

	static int parseNeighborFromRule(String rule) {
		int neighbor = 0;
		int state = 2;
		int i = 0;
		boolean negative = false, getOut = false;

		do {
			for (; i < rule.length(); i++) {
				char letter = rule.charAt(i);

				if (letter >= '0' && letter <= '9')
					neighbor = neighbor * BASE + (letter - '0');
				else if (letter == '-')
					negative = !negative;
				else if (letter == 'N') {
					// really reading state not neighbor
					state = 2;
					neighbor = 0;
				} else {
					getOut = true;
					break;
				}
			}
		} while (state != 2 && neighbor != 0 && !getOut);
		if (neighbor == 0) {
			System.out.println("no neighborhood for rule " +
				rule + " known");
			return Life2DInterface.DEFAULT_NEIGHBOR_INDEX;
		}
		if (state != 2 && state != 4) {
			System.out.println("no state " +
				state + " known");
		}
		if (negative)
			return neighborToIndex(-neighbor);
		return neighborToIndex(neighbor);
	}

	static String parseNeighborRule(String rule) {
		//int state = parseStateFromRule(rule);
		int neighborIndex = parseNeighborFromRule(rule);
		int i = 0, neighbor;

		for (; i < rule.length(); i++) {
			char letter = rule.charAt(i);

			if (!(letter >= '0' && letter <= '9') &&
					letter != '-' &&
					letter != 'N' &&
					!(letter >= 'a' && letter <= 'z'))
				break;
		}
		neighbor = indexToNeighbor[neighborIndex];
		return parseRule(neighbor, rule.substring(i));
	}

	static String parseRule(int neighbors, String rule) {
		int part1, part2;
		StringBuilder[] ruleParts = new StringBuilder[2];
		int i = 0;
		int ni = neighborToIndex(neighbors);
		String defaultString = defaultRuleStrings[ni];

		if (DEBUG)
			System.out.println("parseRule: " + rule);
		ruleParts[0] = new StringBuilder();
		ruleParts[1] = new StringBuilder();
		/*Integer defaultInteger = defaultRuleIndex(ni);
		if (defaultInteger != null)
			defaultString = lifeRules[ni][defaultInteger.intValue()].ruleName;
		if (defaultString == null)*/
		if ("".equals(rule)) {
			if (DEBUG)
				System.out.println("parseRule: null default " +
					defaultString);
			return defaultString;
		}
		switch (rule.charAt(i)) {
		// Fertility
		case 'B':
		case 'F':
			part1 = 0;
			break;
		// Environment
		case 'S':
		case 'E':
			part1 = 1;
			break;
		default:
			{
				if (DEBUG)
					System.out.println("parseRule: 1 default " + defaultString);
				return defaultString;
			}
		}
		i = parsePart(rule, ni, ++i, ruleParts[part1]);
		if (i >= rule.length()) {
			if (DEBUG)
				System.out.println("parseRule: 2 default " + defaultString);
			return defaultString;
		}
		if (rule.charAt(i) == '/') {
			i++;
			if (i >= rule.length()) {
				if (DEBUG)
					System.out.println("parseRule: / default " + defaultString);
				return defaultString;
			}
		}
		switch (rule.charAt(i)) {
		// Fertility
		case 'B':
		case 'F':
			part2 = 0;
			if (part1 == 0) {
				if (DEBUG)
					System.out.println("parseRule: 4 default " + defaultString);
				return defaultString;
			}
			break;
		// Environment
		case 'S':
		case 'E':
			part2 = 1;
			if (part1 == 1) {
				if (DEBUG)
					System.out.println("parseRule: 3 default " + defaultString);
				return defaultString;
			}
			break;
		default:
			{
				if (DEBUG)
					System.out.println("parseRule: 5 default " + defaultString);
				return defaultString;
			}
		}
		parsePart(rule, ni, ++i, ruleParts[part2]);
		if (DEBUG)
			System.out.println("parseRule: " +
				"B" + ruleParts[0].toString() +
				"/S" + ruleParts[1].toString());
		return "B" + ruleParts[0].toString() +
			"/S" + ruleParts[1].toString();
	}

	static int parsePart(String rule, int neighborIndex, int i, StringBuilder sb) {
		char c, oldC;
		boolean gotDash = false;
		int j = -1;
		oldC = '.';
		while (i < rule.length()) {
			c = rule.charAt(i);
			if (c >= 'a' && c <= 'z') {
				if (gotDash && j != -1) {
					ArrayList<Character> letters = new  ArrayList<>();
					letters.add(c);
					for (i++; i < rule.length(); i++) {
						c  = rule.charAt(i);
						if (c >= 'a' && c <= 'z') {
							letters.add(c);
						} else {
							i--;
							break;
						}
					}
					gotDash = false;
					ArrayList<Character> newLetters = invertGroupForChars(groupNumber[neighborIndex][j - 1], letters, indexToNeighbor[neighborIndex]);
					for (char letter: newLetters) {
						sb.append(letter);
					}
				} else
					sb.append(c);
			} else if (c == '-') {
				if (oldC != c) {
					oldC = c;
				}
				gotDash = true;
			} else if (c == '_') {
				if (gotDash) {
					sb.append('-');
					gotDash = false;
				}
				if (oldC != c) {
					sb.append(c);
					oldC = c;
				}
			} else if (Character.isDigit(c)) {
				if (gotDash) {
					sb.append('-');
					gotDash = false;
				}
				if (oldC != c) {
					sb.append(c);
					oldC = c;
				}
				j = c - '0';
			} else {
				break;
			}
			i++;
		}
		return i;
	}

	static String parseRuleFull(String rule) {
		String[] nStrings = rule.split("B");
		String[] sStrings = nStrings[1].split("S");
		String[] bStrings = sStrings[0].split("/");
		String bString = "";
		String sString = "";

		if (sStrings.length > 1)
			sString = sStrings[1];
		if (bStrings.length > 0)
			bString = bStrings[0];
		return nStrings[0] + full("B" + bString) +
			"/" + full("S" + sString);
	}

	static String parseRuleCondense(String rule) {
		String[] nStrings = rule.split("B");
		String[] sStrings = nStrings[1].split("S");
		String[] bStrings = sStrings[0].split("/");
		String bString = "";
		String sString = "";

		if (sStrings.length > 1)
			sString = sStrings[1];
		if (bStrings.length > 0)
			bString = bStrings[0];
		return nStrings[0] + condense("B" + bString) +
			"/" + condense("S" + sString);
	}

	int expand(int number) {
		// find a mask >= 3 where not current in mask but
		// has a "neighbor" that has mask
		int[] ruleMask = new int[Life2DRules.MAX_NEIGHBORS + 1];
		int mask = 1;
		final int lowest = 3;
		int high = lowest;
		int additional = 0;
		int count = 0;

		if (number == 0)
			return 0;
		for (int i = 0; i < Life2DRules.MAX_NEIGHBORS - 1; i++) {
			if ((number & mask) > 0) {
				ruleMask[i] = 1;
				high = i;
			}
			mask <<= 1;
//			System.out.print(ruleMask[i]);
		}
		do {
			additional = generator.nextInt(high + 3 - lowest) + lowest;
//		System.out.println(" " + high + " " + additional + " " + ruleMask[2] + " " + ruleMask[3] + " " + ruleMask[4]);
			count++;
			if (count > 72)
				return number;
			if (additional < 1)
				additional = 1;
			if (additional > Life2DRules.MAX_NEIGHBORS - 1)
				additional = Life2DRules.MAX_NEIGHBORS - 1;
		} while (ruleMask[additional] != 0 ||
			(ruleMask[additional + 1] == 0 &&
			ruleMask[additional - 1] == 0));
//		System.out.println(" " + high + " " + number + " " + additional + " " + (number | (1 << additional)));
		return number | (1 << additional);
	}

	void setStirs() {
		if (generator.nextInt(2) == 0) {
			birthStir = expand(birth);
			survivalStir = survival;
		} else {
			birthStir = birth;
			survivalStir = expand(survival);
		}
	}

	private void setMasks(String fertility, String environment) {
		birth = setMask(full(fertility), birthGroup);
		survival = setMask(full(environment), survivalGroup);
		setStirs();
	}

	public static String toStringGroup(int[] value) {
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < value.length; i++) {
			sb.append(Integer.toString(i)).append(":").append(Integer.toString(value[i]));
			if (i != value.length - 1)
				sb.append(", ");
		}
		return sb.toString();
	}

	// for debugging
	static String toString(int neighbors, Life2DRules param, boolean verbose) {
		int g, l, neighborIndex;
		boolean found;
		StringBuilder sb = new StringBuilder("B");

		if (verbose)
			System.out.print("rule (Birth/Survival " +
				neighbors + " neighborhood): ");
		neighborIndex = neighborToIndex(neighbors);
		for (l = 0; l <= Math.abs(neighbors) && l < BASE; l++) {
			if (param.getBirth(l)) {
				sb.append(l);
			}
			else if (l >= FIRST_GROUP && l < FIRST_GROUP + maxgroups[neighborIndex]) {
				for (g = 0; g < groupNumber[neighborIndex][l - FIRST_GROUP]; g++) {
					if ((param.birthGroup[l - FIRST_GROUP] & (1 << g)) != 0) {
						sb.append(Integer.toString(l)).append('a').append(g);
					}
				}
			}
		}
		sb.append("/S");
		for (l = 0; l <= Math.abs(neighbors) && l < BASE; l++) {
			if (param.getSurvival(l)) {
				sb.append(l);
			}
			else if (l >= FIRST_GROUP && l < FIRST_GROUP + maxgroups[neighborIndex]) {
				for (g = 0; g < groupNumber[neighborIndex][l - FIRST_GROUP]; g++) {
					if ((param.survivalGroup[l - FIRST_GROUP] & (1 << g)) != 0) {
						sb.append(Integer.toString(l)).append('a').append(g);
					}
				}
			}
		}
		found = false;
		for (l = 0; l < maxgroups[neighborIndex]; l++) {
			if ((param.birthGroup[l] != 0) ||
					(param.survivalGroup[l] != 0)) {
				found = true;
				break;
			}
		}
		if (verbose) {
			System.out.println(sb);
			//System.out.println(binary rule :
			//	sb, ((found) ? " (nongroup)" : ""),
			System.out.println("binary rule : Birth 0x" +
				Integer.toHexString(param.birth) +
				", Survival 0x" +
				Integer.toHexString(param.survival));
			if (found) {
				for (l = 0; l < maxgroups[neighborIndex]; l++) {
					System.out.println(
						"groups in neighborhood " +
						(l + FIRST_GROUP) +
						": Birth 0x" +
						Integer.toHexString(param.birthGroup[l]) +
						", Survival 0x" +
						Integer.toHexString(param.survivalGroup[l]));
				}
			}
		}
		return sb.toString();
	}

	static int[] sampleInt6 = {0x3, 0x5, 0x9,
			0x7, 0xb, 0x15,
			0xf, 0x17, 0x1b};
	static char[] sampleChar6 = {'o', 'm', 'p',
			'o', 'm', 'p',
			'o', 'm', 'p'};

	static int[] sampleInt8 = {0x1, 0x2,
			0x5, 0xa, 0x9, 0x3, 0x22, 0x11,
			0x15, 0x2a, 0x29, 0xe, 0x7, 0xd, 0x25, 0x13, 0xb, 0x23,
			0x55, 0xaa, 0x2c, 0xf, 0x36, 0x17, 0x35, 0x39, 0x2b, 0x2e, 0x27, 0x1b, 0x33, 0x2d,
			0xae, 0x57, 0x5b, 0x1f, 0x3e, 0x2f, 0x6b, 0x3b, 0x3d, 0x37,
			0xbe, 0x5f, 0x6f, 0x3f, 0x77, 0xbb,
			0xfe, 0x7f};
	static char[] sampleChar8 = {'c', 'e',
			'c', 'e', 'k', 'a', 'i', 'n',
			'c', 'e', 'k', 'a', 'i', 'n', 'y', 'q', 'j', 'r',
			'c', 'e', 'k', 'a', 'i', 'n', 'y', 'q', 'j', 'r',  't', 'w', 'z', 'f',
			'c', 'e', 'k', 'a', 'i', 'n', 'y', 'q', 'j', 'r',
			'c', 'e', 'k', 'a', 'i', 'n',
			'c', 'e'};

	static int[] sampleInt12 = {0x1, 0x2, 0x4}; // many more values
	static char[] sampleChar12 = {'c', 'e', 'k'}; // experimental, not sure what letters to use yet
	
	static char[] nontotalisticCalc6 = new char[64]; // 2^6
	static char[] nontotalisticCalc8 = new char[256]; // 2^8
	static char[] nontotalisticCalc12 = new char[4096]; // 2^12

	private static int rotateBitsDown(int value, int position, int bits) {
		int mask = (1 << bits) - 1;
		int shiftMask = (1 << position) - 1;
		return ((value & mask) >> position) + ((value & shiftMask) << (bits - position));
	}

	private static int flipBits(int value, int bits) {
		int mask = (1 << bits) - 1;
		int temp = value & mask;
		mask = 1;
		int halfBits = bits >> 1;
		int shiftBits = bits;
		int result = temp & mask;
		for (int i = 1; i < halfBits; i++) {
			mask <<= 1;
			shiftBits -= 2;
			result += (temp >> shiftBits) & mask;
		}
		mask <<= 1;
		shiftBits = 0;
		result += temp & mask;
		for (int i = 1; i < halfBits; i++) {
			mask <<= 1;
			shiftBits += 2;
			result += (temp << shiftBits) & mask;
		}
		return result;
	}

	private static void calculateTot(int bits) {
		for (int i = 0; i < (1 << bits); i++) {
			if (bits == 6)
				nontotalisticCalc6[i] = 'x';
			else if (bits == 8)
				nontotalisticCalc8[i] = 'x';
			else if (bits == 12)
				nontotalisticCalc12[i] = 'x';
		}
		for (int s = 0; s < ((bits == 6) ? sampleInt6.length : ((bits == 8) ? sampleInt8.length : sampleInt12.length)); s++) {
			int k = (bits == 6) ? sampleInt6[s] : ((bits == 8) ? sampleInt8[s] : sampleInt12[s]);

			for (int j = 0; j < ((bits == 6) ? (bits << 1) : ((bits == 8) ? bits : (bits >> 1))); j++) {
				if (j == ((bits == 6) ? bits : ((bits == 8) ? (bits >> 1) : (bits >> 2)))) {
					k = flipBits(k, bits);
				} else {
					k = ((bits == 6) ? rotateBitsDown(k, 1, bits) : ((bits == 8) ? rotateBitsDown(k, 2, bits) : rotateBitsDown(k, 4, bits)));
				}
				if (bits == 6)
					nontotalisticCalc6[k] = sampleChar6[s];
				else if (bits == 8)
					nontotalisticCalc8[k] = sampleChar8[s];
				else
					nontotalisticCalc12[k] = sampleChar12[s];
			}	
		}
		for (int i = 0; i < (1 << bits); i++) {
			char letter = ((bits == 6) ? nontotalisticCalc6[i] : ((bits == 8) ? nontotalisticCalc8[i] : nontotalisticCalc12[i]));
			System.out.print("'" + letter + "', ");
			if ((i & ((1 << (bits >> 1)) - 1)) == (1 << (bits >> 1)) - 1) {
				System.out.println();
				if (i == (1 << bits) - 1) {
					System.out.println();
				}
			}
		}
	}

	public static void main(String[] args) {
		calculateTot(6);
		calculateTot(8);
		calculateTot(12);
		readRulesXML(Life2DInterface.XML_FILE, null, true);
		readRulesXML(Life2DInterface.XML_FILE, null, false);
	}
}
